package no.rwr.xutil;

import no.rwr.x.*;
//import no.rwr.render.*;
import java.awt.image.*;
import java.awt.Point;
import java.awt.Transparency;
import java.awt.color.*;

/** 
    Encapsulates both XImage and Pixmap in a common abstraction. One of
    the benefits of this is that it will allow a natural handling of
    SHM. 

    This class also acts as bridge between the no.rwr.xutil and
    no.rwr.render packages.


    Maybe this class should be generic and have a subclass that is
    tied to X. In that case, maybe the best place for the generic
    class would be no.rwr.render.  */

public class Image {
    public XImage ximage;
    Visual visual;
    XImageFactory factory;
    
    Pixmap pixmap;
    boolean needFlush = true;
    DataBuffer dataBuffer;
    
    /* FIXME: Change constructors, either to take an existing XImage
       or an XImageFactory. */

    public Image(XImageFactory factory, Raster raster) {
	this.visual = factory.getVisual();
	this.factory = factory;
	ximage = factory.createXImage(raster);
    }
    

    public Image(Visual visual, int width, int height) {
	this.visual = visual;

	/* FIXME: we probably don't want XImage to auto-allocate the
	   memory, since we want to be able to construct a
	   WritableRaster from it. Problem: in order to allocate an
	   array of the right data type, we need to know which color
	   model we are going to use. The problem is that we have not
	   identified a color model at this time. Maybe we should:

	   1. Create ximage.

	   2. Create factory based on ximage.

	   3. Create array based on factory.

	   4. Set array as data for ximage.

	   This is an expensive road to take, so we encourage the use
	   of the factory when creating new ximages.

	   The key point here is that between step 1 and step 4 we
	   need to determine the data type to use. This should
	   probably be done in XImageFactory at 2.  */

	ximage = new XImage(visual, width, height,
			    false // do not auto-allocate memory
			    );
	factory = new XImageFactory(ximage, visual);
	dataBuffer = factory.getSampleModel().createDataBuffer();
	XImageFactory.attachData(ximage, dataBuffer, 0);
    } 


    public WritableRaster getWritableRaster() {
	return new WritableRaster(factory.getSampleModel(), 
				  dataBuffer,
				  new Point(0, 0));
    }

    public Object getData() {
	return ximage.getData();
    }

    public int getBytesPerLine() {
	return ximage.getBytesPerLine();
    }
    
    public Pixmap getPixmap() {
	if (pixmap == null) {
	    pixmap = new Pixmap(visual.getScreen().getRootWindow(),
				ximage.getWidth(), ximage.getHeight(),
				ximage.getDepth());
	}
	if (needFlush) {
	    // FIXME: don't create a new GC all the time.
	    GC gc = new GC(pixmap);
	    gc.putImage(ximage,
			0, 0,
			0, 0,
			ximage.getWidth(), ximage.getHeight());
	    needFlush = false;
	}
	return pixmap;
    }

    /** This get the format of the image in the no.rwr.render sense of
        the word. This method should not generally be used. The
        intended image format usually be determined before an Image
        object is created. Usually a source and target format and a
        renderer should be determined, and used throughout the whole
        program. Reidentifiying the correct renderer can occur some
        overhead. Especially for renderers that initializes tables
        used for rendering. */
    /*       
    public ImageFormat getImageFormat() {

	// FIXME: assumes True Color

	return new ImageFormat(ximage.getBitsPerPixel(),
			       ximage.getRedMask(),
			       ximage.getGreenMask(),
			       ximage.getBlueMask());
    }
    */

    /** 
	@deprecated use getSampleModel() instead

    */
    public SampleModel getSampleFormat() {
	return getSampleModel();
    }


    /* FIXME: It is probably better to find a ColorModel instead,
       since we don't want to want to bind to a specific width height,
       and we want to store sample interpretation, ie. number of bits
       per subfield. */

    /* NO! We need both ColorModel and SampleModel:

       ColorModel is needed to describe interpretation of pixels,
       eg. scaling or lookup operation.

       SampleModel is needed to describe the format of the pixels,
       eg. band order, format, (padding). Problem is that sample model
       is bound to a specific size.

       Solution is to create an equivalent to GraphicsConfiguration
       that can give us both. ...or just give us ColorModel, but be
       capable of creating a raster (BufferedImage equivalent?) on
       demand. The color model can be extracted from the raster
       afterwards.

    */


    /* To create a ColorModel:

       + get masks from Visual

       + somehow find depth from visual (Visual is opaque) (use
         VisualInfo?)

       + bits per pixel is a determined by display and depth

       + get bits per pixel from ScreenFormat stored in display
         (dpy->pixmap_format)
    
       + if no suitable screen format exists, round depth up to 4, 8,
         16, or 32
    
    */
    
    public static ColorModel getColorModel(Visual v) {

	int bitsPP = 24; // FIXME: get this from ScreenFormat
	
    tryByteComponentModel:
	if ((bitsPP%8) == 0) {
	    System.out.println("pixels are byte multiples"); // DEBUG
	    
	    
	    int ri = getLSBFirstByteOffsetFromMask(v.getRedMask());
	    if (ri == -1) break tryByteComponentModel;
	    int gi = getLSBFirstByteOffsetFromMask(v.getGreenMask());
	    if (gi == -1) break tryByteComponentModel;
	    int bi = getLSBFirstByteOffsetFromMask(v.getBlueMask());
	    if (bi == -1) break tryByteComponentModel;
	    System.out.println("samples are bytes"); // DEBUG

	    int bytesPP = bitsPP / 8;
	    
	    boolean reverseByteOrder = false; // FIXME: don't assume this
	    if (reverseByteOrder) {
		ri = bytesPP - ri;
		gi = bytesPP - gi;
		bi = bytesPP - bi;
	    }
	    int[] bandOffsets = { ri, gi, bi };

	    ColorSpace sRGBCS =
		ColorSpace.getInstance(ColorSpace.TYPE_sRGB);
	    int[] bits = { 8, 8, 8 };

	    return new ComponentColorModel(sRGBCS,
					   bits,
					   false, // does not have alpha
					   false, // no premultiplied alpha
					   Transparency.OPAQUE,
					   DataBuffer.TYPE_BYTE);
	}
	
	throw new UnsupportedOperationException("Don't know how to " +
						"create color model" +
						" for visual " + v);
	    
	

    }
    

    public XImageFactory getXImageFactory() {
	if (factory == null) {
	    factory = new XImageFactory(ximage, visual);
	}
	
	return factory;
    }
    
    

    public SampleModel getSampleModel() {
	
	if (factory != null) // FIXME: make unconditional
	    return getXImageFactory().getSampleModel();

	/* FIXME: Transform the following code into the creation of a
	   new factory. */

	// FIXME: assumes true or direct color (ie. visual with subfields)

	/* Problem here: should we always use TYPE_BYTE or should we
           adjust type to match XImage unit? This will create problems
           related to endianess. */

	System.out.println("Creating sample model from" +  // DEBUG
			   ximage);

        tryComponentModel:
	if (ximage.getDepth() == 24) {
	    int bitsPP = ximage.getBitsPerPixel();
	    if ((bitsPP % 8) != 0) break tryComponentModel;
	    System.out.println("pixels are byte multiples"); // DEBUG
	    
	    int ri = getLSBFirstByteOffsetFromMask(ximage.getRedMask());
	    if (ri == -1) break tryComponentModel;
	    int gi = getLSBFirstByteOffsetFromMask(ximage.getGreenMask());
	    if (gi == -1) break tryComponentModel;
	    int bi = getLSBFirstByteOffsetFromMask(ximage.getBlueMask());
	    if (bi == -1) break tryComponentModel;
	    System.out.println("samples are bytes"); // DEBUG

	    int bytesPP = bitsPP / 8;
	    
	    boolean reverseByteOrder = false; // FIXME: don't assume this
	    if (reverseByteOrder) {
		ri = bytesPP - ri;
		gi = bytesPP - gi;
		bi = bytesPP - bi;
	    }
	    int[] bandOffsets = { ri, gi, bi };
	    
	    return new ComponentSampleModel(DataBuffer.TYPE_BYTE,
					    ximage.getWidth(),
					    ximage.getHeight(),
					    bytesPP,
					    ximage.getBytesPerLine(),
					    bandOffsets);
	}

	int dataType = 0;
	int bytesPerElement = 0;

	// FIXME: This is not an useful way to determine type.
	switch (ximage.getBitsPerPixel()) {
	case 16:
	    dataType = DataBuffer.TYPE_USHORT;
	    bytesPerElement = 2;
	    break;
	case 8:
	    dataType = DataBuffer.TYPE_BYTE;
	    bytesPerElement = 1;
	    break;
	default:
	    throw new InternalError();
	}

	int[] bitMasks = { ximage.getRedMask(),
			   ximage.getGreenMask(),
			   ximage.getBlueMask() };

	int elements = ximage.getBytesPerLine()/bytesPerElement;

	return new
	    SinglePixelPackedSampleModel(dataType,
					 ximage.getWidth(),
					 ximage.getHeight(),
					 elements,
					 bitMasks);
    }

    private static int getLSBFirstByteOffsetFromMask(int mask) {
	switch (mask) {
	case 0xff:
	    return 0;
	case 0xff00:
	    return 1;
	case 0xff0000:
	    return 2;
	case 0xff000000:
	    return 3;
	default:
	    // FIXME: maybe throw exception instead?
	    return -1; 
	}
    }
    


    /* One thing we shouldn't do in these rutines: Worry about bit-
       and byte-ordering. The magic in Xlib's XPutImage does this.

       Byte and bit ordering usually is not a needed on local servers,
       and if the data is pushed to a remote server, then bandwith is
       usually more of a problem than the slowness of converting.  */




    /* Dither:

       abcdefgh
-           abc      remove to compensate for adding dither
       rrrrrrrr
+           ddd      add dither
       RRRRR...

       11111111      No effect on pure colors
-           111
       11111000
       --------
            000      d000
       11111000      ---> 11111
       --------
            111      d111
       11111000      ---> 11111

       10101010
-           101
       10100101
       --------
            000      d000
       10100101      ---> 10100
       --------
            111      d111
       10101100      ---> 10101

       00000000
-           000
       00000000
+           111
       00000111     ---> 00000


       alt:

       11111111
+           111      d111
      100000110
-          1000
      011111110
       =====
       
       11111111
+           000      d000
      011111111      ---> 11111
-          0111
      011111000      ---> 11111
       =====      

       00000000
+           111
      000000111
-          0000
      000000111     ---> 00000
       =====

       v
       v += dither3[xy];
       v -= v>>>(outBits-1);
       v >>>= discardBits



       v
       v -= v>>>outBits;
       v += dither3[xy];
       v >>>= discardBits;

                     77776666555544443333222211110000
       0 4 1 5                  # # # # # # # #######
       6 2 7 3         # # # # # # # # ##############
       1 5 0 4                #   # # # # # ### #####
       7 3 6 2       #   # # # # # # ### ############
       0 4 1 5                  # # # # # # # #######
       6 2 7 3         # # # # # # # # ##############
       1 5 0 4                #   # # # # # ### #####
       7 3 6 2       #   # # # # # # ### ############
       0 4 1 5                  # # # # # # # #######
       6 2 7 3         # # # # # # # # ##############
       1 5 0 4                #   # # # # # ### #####
       7 3 6 2       #   # # # # # # ### ############
    */

}
